/*
 * Author: Utkarsh Gupta
 * Created on: June 5, 2012
 * File Name : refinement2.cpp
 * Project Name : Automation of Folding (DDP)
 * Description : This file creates the structural refinement
 *            for the Bipartite PG Graph. A network of PPUs, PMUs
 *            and wire connections betwwen them representing the PG Graph
 */

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <malloc.h>
#include "C_Functions/data_structs.h"
#include "C_Functions/auxiliary_functions.h"

using std::vector;
using namespace std;

void refinement2( parameters &r_1)
{

    int **p_edge_num;    // IN THE FOLDED PG GRAPH
                                        // p_edge_num[ <Folded Point Node Index> ][ <Edge Number> ] = Folded Hyp Node Index
    int **h_edge_num;    // IN THE FOLDED PG GRAPH
                                        // h_edge_num[ <Folded Hyp Node Index> ][ <Edge Number> ] = Folded Point Node Index


    p_edge_num = new int *[r_1.folded_nodes];
    h_edge_num = new int *[r_1.folded_nodes];
    for (int i = 0; i < r_1.folded_nodes; ++i){
        p_edge_num[i] = new int [r_1.num_p_edges];
        h_edge_num[i] = new int [r_1.num_h_edges];
    }

    int edge_num = 0; // Edge Counter
                      // Increases from 0 to r_1.num_h_edges-1

    int two_edge_counter = 0; // Used incase of 2 connections from H0 to a single FOLDED POINT NODE
                              // This is the case when r_1.h0_to_p[<Folded Point Index>] = 2
                              // Can not be greater than 2 as maximum possible accesses in 1 cycle are 2

    int point_node_num = 0; // Indicates the current point index
                            // Increases from 0 to r_1.folded_nodes - 1

    //Initializing Hyperplane #0 Edges to their respective POINT Connections
    while (point_node_num < r_1.folded_nodes){
        if(r_1.h0_to_p[point_node_num] > 0) // If H0 has 1 or more Connections with
                                        // Point Node *point_node_num*
                                        // then assign current edge to this Point Node
        {
            h_edge_num[0][edge_num] = point_node_num;

            if(r_1.h0_to_p[point_node_num] == 1){    // In case of just 1 connection
                point_node_num++;                // Move forward by increasing the point_node_num
            }
            else{
                two_edge_counter++;        // In case of more than 1 edge i.e. 2 Connections
                                        // Increment two_edge_counter by 1 and not the point_node_num
                                        // After writing the FIRST Connection out of the 2
                                        // two_edge_counter will be 1
                                        // After writing the SECOND Connection out of the 2
                                        // two_edge_counter will increment to 2
            }
            edge_num++;        // Move to the next EDGE by incrementing edge_num
        }
        else{
            point_node_num++;
        }

        if ( two_edge_counter == 2){    // After writing the FIRST Connection out of the 2
                                        // two_edge_counter will be 1 and will not enter this loop
                                        // After writing the SECOND Connection out of the 2
                                        // we must now increment the point_node_num
            point_node_num++;
            two_edge_counter = 0;        // For the next point_node_num the two_edge_counter
                                        // must return to its intial state that is *ZERO*
        }
    }

    // Declaring Edges for all other Hyp Nodes ( #1 to #<r_1.folded_nodes - 1> )
    for (int h_node = 1; h_node < r_1.folded_nodes; h_node++){
        for (int edge_num = 0; edge_num < r_1.num_h_edges; edge_num++){
        h_edge_num[h_node][edge_num] = (h_edge_num[0][edge_num] + h_node) % r_1.folded_nodes;
                                                // Rotated in circulant fashion
        }
        sort2(h_edge_num[h_node], r_1.num_h_edges);    // All edge connections for HYP #<h_node>
                                                // have been generated in circulant fashion
                                                // simple *Linear* sorting would not change the
                                                // circulancy but only the order in which we list edges
                                                // i.e. for each Hyp node connections starting from POINT NODE #0
    }

    // Re-declaring these variables to *ZERO* for use in
    // writing the Point Node Edge Connection Array
    edge_num = 0;
    two_edge_counter = 0;

    int hyp_node_num = 0;    // Indicates the current hyperplane index
                            // Increases from 0 to r_1.folded_nodes - 1

    //Declaring Point #0 Edges to their respective HYPERPLANE Connections
    while (hyp_node_num < r_1.folded_nodes)        // A similar algorithm as above is used to write the
                                            // Point Node Connection Array
    {
        if(r_1.p0_to_h[hyp_node_num] > 0){
            p_edge_num[0][edge_num] = hyp_node_num;
            if(r_1.p0_to_h[hyp_node_num] == 1){
                hyp_node_num++;
            }
            else{
                two_edge_counter++;
            }
            edge_num++;
        }
        else{
            hyp_node_num++;
        }

        if ( two_edge_counter == 2){
            hyp_node_num++;
            two_edge_counter = 0;
        }
    }

    // Declaring Edges for all other Point Nodes ( #1 to #<r_1.folded_nodes - 1> )
    for (int p_node = 1; p_node < r_1.folded_nodes; p_node++){
        for (int edge_num = 0; edge_num < r_1.num_p_edges; edge_num++){
            p_edge_num[p_node][edge_num] = (p_edge_num[0][edge_num] + p_node) % r_1.folded_nodes;
        }
        sort2(p_edge_num[p_node], r_1.num_p_edges); // Connections are written in circulant manner
                                               // though sorted linearly to ensure connections are in
                                               // order starting from the ones with H0
    }

    for (int i = 0; i < r_1.folded_nodes; i++){
        for (int j = 0; j < r_1.num_h_edges; j++){
            cout << h_edge_num[i][j] << "  ";
            }
        cout << endl;
    }

    cout << endl;

    for (int i = 0; i < r_1.folded_nodes; i++){
        for (int j = 0; j < r_1.num_p_edges; j++){
            cout << p_edge_num[i][j] << "  ";
            }
        cout << endl;
    }

    int hyp_edges = r_1.folded_nodes * r_1.num_h_edges; // Total number of HYPERPLANE EDGES
    int point_edges = r_1.folded_nodes * r_1.num_p_edges; // Total number of POINT EDGES

      // Instantiating Hyperplane and Point PPUs
    r_1.H_PPU  = (PPU*)malloc(r_1.folded_nodes*sizeof(PPU));
    r_1.P_PPU = (PPU*)malloc(r_1.folded_nodes*sizeof(PPU));

    // Instantiating Hyperplane and Point PMUs
    r_1.H_PMU = (PMU*)malloc(r_1.folded_nodes*sizeof(PMU));
    r_1.P_PMU = (PMU*)malloc(r_1.folded_nodes*sizeof(PMU));


    // Initializing H/P PPUs and PMUs
    for (int node_num = 0; node_num < r_1.folded_nodes; node_num++){
        // Looking from the Hyperplane side
        // Each of r_1.H_PPU and r_1.P_PMU have num_p_edge connections (ports)
        r_1.P_PMU[node_num].id = node_num,
        r_1.P_PMU[node_num].num_ops = r_1.num_h_edges;
        r_1.P_PMU[node_num].op_ports = new int[r_1.num_h_edges];

        r_1.H_PPU[node_num].id = node_num,
        r_1.H_PPU[node_num].num_ips = r_1.num_h_edges;
        r_1.H_PPU[node_num].ip_ports = new int[r_1.num_h_edges];
        // Looking from the Point side
        // Each of r_1.H_PMU and r_1.P_PPU have num_h_edge connections (ports)
        r_1.H_PMU[node_num].id = node_num,
        r_1.H_PMU[node_num].num_ops = r_1.num_p_edges;
        r_1.H_PMU[node_num].op_ports = new int[r_1.num_p_edges];

        r_1.P_PPU[node_num].id = node_num,
        r_1.P_PPU[node_num].num_ips = r_1.num_p_edges;
        r_1.P_PPU[node_num].ip_ports = new int[r_1.num_p_edges];
    }

    //Wire Channel for the FOLDED PG GRAPH
    r_1.h_to_p_wires  = (wire*)malloc(point_edges*sizeof(wire)); // Hyperplane PHYSICAL_MEM_UNIT to Point PYSICAL_PROC_UNIT
    r_1.p_to_h_wires  = (wire*)malloc(hyp_edges*sizeof(wire)); // Point PHYSICAL_MEM_UNIT to Hyperplane PYSICAL_PROC_UNIT

    /* We need 2 sets for wires:
            1) First set of connections between r_1.H_PPU and r_1.P_PMU : p_to_h_wires
            2) Second set of connections between r_1.P_PPU and r_1.H_PMU :  h_to_p_wires

    These set of connections may be different if any set contains 2 edges between the
    some 2 nodes (one on either side). Either between an r_1.H_PPU and a r_1.P_PMU or a r_1.P_PPU
    and an r_1.H_PMU or both.
    */

    int *op_port_ctr_H = new int[r_1.folded_nodes];     // Array Index is the Folded Point Node Index
                                        // The array element for each point node from #0 to
                                        // #<r_1.folded_nodes - 1> is a counter of the number of
                                        // ports occupied

    // Initializing occupied ports for all HYP NODES as *ZERO*
    for (int hyp_id = 0; hyp_id < r_1.folded_nodes; hyp_id++){
        op_port_ctr_H[hyp_id] = 0;
    }
    cout << "\nbefore h->p\n";
    // Declaring the elements for *all* H->P Wires
    for (int h_edge = 0; h_edge < point_edges; h_edge++){

        r_1.h_to_p_wires[h_edge].id = h_edge;

        r_1.h_to_p_wires[h_edge].PPU_id = (h_edge - (h_edge % r_1.num_p_edges))/ r_1.num_p_edges;

        r_1.h_to_p_wires[h_edge].PPU_port_num = h_edge % r_1.num_p_edges;

        r_1.P_PPU[ r_1.h_to_p_wires[h_edge].PPU_id ].ip_ports[ r_1.h_to_p_wires[h_edge].PPU_port_num ] = h_edge;

        r_1.h_to_p_wires[h_edge].PMU_id = p_edge_num [ (r_1.h_to_p_wires[h_edge].PPU_id) ] [ (r_1.h_to_p_wires[h_edge].PPU_port_num) ];

        r_1.h_to_p_wires[h_edge].PMU_port_num = op_port_ctr_H[r_1.h_to_p_wires[h_edge].PMU_id];

        op_port_ctr_H[r_1.h_to_p_wires[h_edge].PMU_id] ++;

        r_1.H_PMU[r_1.h_to_p_wires[h_edge].PMU_id].op_ports[r_1.h_to_p_wires[h_edge].PMU_port_num] = h_edge;
    }

    int *op_port_ctr_P = new int[r_1.folded_nodes];

    // Initializing occupied ports for all POINT NODES as *ZERO*
    for (int point_id = 0; point_id < r_1.folded_nodes; point_id++){
        op_port_ctr_P[point_id] = 0;
    }
    cout << "\nbefore p->h\n";
    // Declaring the elements for *all* P->H Wires
    for (int p_edge = 0; p_edge < hyp_edges; p_edge++){
        r_1.p_to_h_wires[p_edge].id = p_edge;    // P->H Wire ID. May be redundant.

        // H->P Wire is connected between r_1.P_PMU and r_1.H_PPU
        r_1.p_to_h_wires[p_edge].PPU_id = (p_edge - (p_edge % r_1.num_h_edges))/ r_1.num_h_edges;
                                    // Each r_1.P_PMU will have r_1.num_h_edges (looking from point side)
                                    // As p_edge goes from 0 to point_edges ( r_1.folded_nodes*r_1.num_h_edges )
                                    // PMU ID should go from 0 to r_1.folded_nodes - 1

        r_1.p_to_h_wires[p_edge].PPU_port_num = p_edge % r_1.num_h_edges;
                                    // PMU Port number should increase from 0 to r_1.num_h_edges - 1

        r_1.H_PPU[r_1.p_to_h_wires[p_edge].PPU_id].ip_ports[r_1.p_to_h_wires[p_edge].PPU_port_num] = p_edge;
                                    // P->H Wire *p_edge* is also connected to r_1.P_PMU ID as defined above (say "p")
                                    // at Port as defined above (say "y")
                                    // So, Output port #y of r_1.P_PMU #p is connected to P->H Wire *p_edge*

        r_1.p_to_h_wires[p_edge].PMU_id = h_edge_num [ (r_1.p_to_h_wires[p_edge].PPU_id) ] [ (r_1.p_to_h_wires[p_edge].PPU_port_num) ];
                                    // r_1.H_PMU ID can be obtained from the array p_edge_num[h][x]
                                    //                                h -- r_1.H_PPU ID
                                    //                                x -- r_1.H_PPU Port number

        r_1.p_to_h_wires[p_edge].PMU_port_num = op_port_ctr_P[r_1.p_to_h_wires[p_edge].PMU_id];
                                    // On declaring an edge to r_1.H_PPU ID as described above (say "p")
                                    // the Port Number for that PPU will be op_port_ctr_P [p]

        op_port_ctr_P[r_1.p_to_h_wires[p_edge].PMU_id] ++; // For r_1.H_PPU #p increment occupied port

        r_1.P_PMU[r_1.p_to_h_wires[p_edge].PMU_id].op_ports[r_1.p_to_h_wires[p_edge].PMU_port_num] = p_edge;
                                    // P->H Wire *p_edge* is connected to r_1.H_PPU ID as defined above (say "h")
                                    // at Port as defined above (say "x")
                                    // So, Input port #x of r_1.H_PPU #h is connected to P->H Wire *p_edge*
    }

    return;
    // Refinement 2 is complete here.
}


