/*
 * Author: Utkarsh Gupta
 * Created on: June 5, 2012
 * File Name : make_mem_module.cpp
 * Project Name : Automation of Folding (DDP)
 * Description : This file creates the Memory Module VHDL File
 *            along with its component modules.
 */

#include <iostream>
#include <fstream>
#include <cmath>
#include <string.h>
#include <vector>
#include "C_Functions/auxiliary_functions.h"

using std::vector;
using namespace std;

void make_tdp_mem(int order, int fold_factor, const char* outputfile){

    // Writing the file Generated_Files/Portions/TDP_Mem/tdp_mem2.vhd
    // This portion of the file contains the number of RAM locations,
    // which is dependant on the ORDER and FOLD FACTOR for the
    // PG Bipartite Graph
    ofstream outfile ("Generated_Files/Portions/TDP_Mem/tdp_mem2.vhd");
    outfile << "TYPE RAM_TYPE IS ARRAY ("<< (fold_factor * (order + (order%2))) <<" DOWNTO 0) OF STD_LOGIC;\n\t\t";
    outfile << "SIGNAL ram : RAM_TYPE;\n";
    outfile << "BEGIN \nPROCESS( clk, rst ) \nBEGIN \n\tIF( rst = '1' ) THEN\n\t\tL1: FOR i in 0 to ";
    outfile <<(fold_factor * (order + (order%2))) << " LOOP\n";
    outfile.close();

    outfile.open( outputfile );    // Creating the outputfile
    outfile.close();

    // Appending the tdp_mem1/2/3.vhd together to make tdp_mem.vhd
    append( "HDL_partitioned_files/TDP_Mem/tdp_mem1.vhd", outputfile );
    append( "Generated_Files/Portions/TDP_Mem/tdp_mem2.vhd", outputfile );
    append( "HDL_partitioned_files/TDP_Mem/tdp_mem3.vhd", outputfile );

    return;
}

void make_add_mux(int order, int fold_factor, string default_addr, const char* outputfile){

    // Writing the file Generated_Files/Portions/Addr_Mux/add_mux2.vhd
    ofstream outfile ("Generated_Files/Portions/Addr_Mux/add_mux2.vhd");
    outfile << " <= \"" << default_addr;
    outfile.close();

    outfile.open(outputfile);    // Creating the file
    outfile.close();                            // add_mux.vhd

    // Appending the add_mux1/2/3.vhd together to make add_mux.vhd
    append( "HDL_partitioned_files/Addr_Mux/add_mux1.vhd", outputfile );
    append( "Generated_Files/Portions/Addr_Mux/add_mux2.vhd", outputfile );
    append( "HDL_partitioned_files/Addr_Mux/add_mux3.vhd", outputfile );

    return;
}

void make_rd_addrgen(int order, int fold_factor, string default_addrZ, string default_addr0, int mem_loc, const char* outputfile){

    // Writing the file rd_addrgen2/4/7.vhd
    ofstream outfile ("Generated_Files/Portions/Rd_addrgen/rd_addrgen2.vhd");
    ofstream outfile2 ("Generated_Files/Portions/Rd_addrgen/rd_addrgen4.vhd");
    ofstream outfile3 ("Generated_Files/Portions/Rd_addrgen/rd_addrgen7.vhd");

    outfile << "\ta1: addr1     <= \""<< default_addrZ << "\";\n";
    outfile << "\ta2: addr2     <= \""<< default_addrZ << "\";\n";

    outfile2 << "\n\t\t\taddr1     <= CONV_STD_LOGIC_VECTOR( count1, "<< mem_loc << " );\n";
    outfile2 << "\t\t\taddr2     <= CONV_STD_LOGIC_VECTOR( count2, "<< mem_loc << " );\n";
    outfile3 << "\n\t\taddr1     <= \""<< default_addr0 << "\";\n";
    outfile3 << "\t\taddr2     <= \""<< default_addr0 << "\";\n";

    outfile3.close();
    outfile2.close();
    outfile.close();

    outfile.open(outputfile);    // Creating the file
    outfile.close();                            // rd_addrgen.vhd

    // Appending the rd_addrgen1-8.vhd together to make rd_addrgen.vhd
    append( "HDL_partitioned_files/Rd_addrgen/rd_addrgen1.vhd", outputfile );
    append( "Generated_Files/Portions/Rd_addrgen/rd_addrgen2.vhd", outputfile );
    append( "HDL_partitioned_files/Rd_addrgen/rd_addrgen3.vhd", outputfile );
    append( "Generated_Files/Portions/Rd_addrgen/rd_addrgen4.vhd", outputfile );
    append( "HDL_partitioned_files/Rd_addrgen/rd_addrgen5.vhd", outputfile );
    append( "Generated_Files/Portions/Rd_addrgen/rd_addrgen4.vhd", outputfile );
    append( "HDL_partitioned_files/Rd_addrgen/rd_addrgen6.vhd", outputfile );
    append( "Generated_Files/Portions/Rd_addrgen/rd_addrgen7.vhd", outputfile );
    append( "HDL_partitioned_files/Rd_addrgen/rd_addrgen8.vhd", outputfile );

    return;
}

void make_wr_addrgen(int order, int nodes, int fold_factor, int hyp_node[], string default_addrZ, int mem_loc, const char *outputfile1, const char *outputfile2)
{
    //Please read "For understanding WRITE ALGO.xlsx" for this portion
    #ifndef WRITE_ALGORITHM
    #define WRITE_ALGORITHM
        int even_order = order + (order % 2);

        vector<vector<int> > hyp_conn;
        vector<vector<int> > poi_conn;

        hyp_conn.resize(nodes);
        poi_conn.resize(nodes);
        for (int i = 0; i < nodes; ++i){
            hyp_conn[i].resize(even_order);
            poi_conn[i].resize(even_order);
        }

        for (int x = 0 ; x < order ; x++){
            hyp_conn[0][x] = hyp_node[x];
            if(hyp_conn[0][x]%nodes == 0){
                  poi_conn[0][x] =  hyp_conn[0][x];
            }
            else{
                  poi_conn[0][x] =  nodes - hyp_conn[0][x] ;
            }
        }

        // You need to sort the Hyperplane and Point node connections
        int temp;
        for (int a = 0 ; a < order ; a++){
            for (int x = 0 ; x < order-1 ; x++){
                if(poi_conn[0][x] > poi_conn[0][x+1]){
                      temp = poi_conn[0][x];
                      poi_conn[0][x] = poi_conn[0][x+1];
                      poi_conn[0][x+1] = temp;
                }
            }
        }

        if(order < even_order ){
                hyp_conn[0][even_order-1]= -1;
                poi_conn[0][even_order-1]= -1;
            }

        for (int y = 1 ; y < nodes ; y++){
            for (int z = 0; z < even_order-1 ; z++){
                hyp_conn[y][z]=(hyp_conn[0][z]+y)%nodes;
                poi_conn[y][z]=(poi_conn[0][z]+y)%nodes;
            }
            if(order < even_order ){
                hyp_conn[y][even_order-1]= -1;
                poi_conn[y][even_order-1]= -1;
            }
            else {
                hyp_conn[y][even_order-1]=(hyp_conn[0][even_order-1]+y)%nodes;
                poi_conn[y][even_order-1]=(poi_conn[0][even_order-1]+y)%nodes;
            }
        }

        int nodes_per_fold = nodes/fold_factor;

        vector<vector<vector<int> > > write_sch_mem_poi;
        vector<vector<vector<int> > > write_sch_mem_hyp;

        write_sch_mem_poi.resize(nodes_per_fold);
        write_sch_mem_hyp.resize(nodes_per_fold);
        for (int i = 0; i < nodes_per_fold; ++i) {
            write_sch_mem_poi[i].resize(2);
            write_sch_mem_hyp[i].resize(2);
            for (int k = 0; k < 2; k++){
                write_sch_mem_poi[i][k].resize(even_order * fold_factor);
                write_sch_mem_hyp[i][k].resize(even_order * fold_factor);
            }
        }

        for(int a = 0; a < nodes_per_fold; a++){
            for(int c = 0; c<even_order/2; c++){
                for(int b = 0; b < fold_factor; b++){
                    write_sch_mem_hyp[a][0][ 2*b + (2*c*fold_factor)] = hyp_conn[(b*nodes_per_fold)+a][2*c];
                    write_sch_mem_hyp[a][0][ 2*b + (2*c*fold_factor) + 1] = hyp_conn[(b*nodes_per_fold)+a][2*c +1];
                    write_sch_mem_poi[a][0][ 2*b + (2*c*fold_factor)] = poi_conn[(b*nodes_per_fold)+a][2*c];
                    write_sch_mem_poi[a][0][ 2*b + (2*c*fold_factor) + 1] = poi_conn[(b*nodes_per_fold)+a][2*c +1];
                }
            }
        }

        int hyp, point;
        int to_search_poi, to_search_hyp;
        int edge_no;

        vector<vector<int> > conn;
        conn.resize(nodes);
        for (int i = 0; i < nodes; ++i){
            conn[i].resize(even_order);
        }

        for(int c = 0; c < even_order/2; c++){
            for(int b = 0; b < fold_factor; b++){
                for (int a = 0; a < nodes_per_fold; a++){
                    conn[(b*nodes_per_fold) + a][2*c] = 2*(c*fold_factor + b);
                    conn[(b*nodes_per_fold) + a][2*c + 1] = 2*(c*fold_factor + b) + 1;
                }
            }
        }


        for (int a = 0; a < nodes_per_fold; a++){
            for (int x = 0; x < (even_order*fold_factor); x++){
                point = write_sch_mem_hyp[a][0][x];
                to_search_poi = (((x % (2*fold_factor))/2)*nodes_per_fold) + a ;
                if(point >= 0){
                    edge_no = 0;
                    while(poi_conn[point][edge_no] != to_search_poi)
                        edge_no++;
                    write_sch_mem_hyp[a][1][x] = conn[point][edge_no];
                }
                else{
                    write_sch_mem_hyp[a][1][x] = conn[to_search_poi][(2*(x/(2*fold_factor))) + (x%2)];
                }

                hyp = write_sch_mem_poi[a][0][x];
                to_search_hyp = (((x % (2*fold_factor))/2)*nodes_per_fold) + a ;
                if(hyp >= 0){
                    edge_no = 0;
                    while(hyp_conn[hyp][edge_no] != to_search_hyp)
                        edge_no++;
                    write_sch_mem_poi[a][1][x] = conn[hyp][edge_no];
                }
                else{
                    write_sch_mem_poi[a][1][x] = conn[to_search_hyp][(2*(x/(2*fold_factor))) + (x%2)];
                }
            }
        }
    #endif

    ofstream outfile ("Generated_Files/Portions/Wr_addrgen/hyp_wr_add2.vhd");
    ofstream outfile2 ("Generated_Files/Portions/Wr_addrgen/point_wr_add2.vhd");

    for (int nodes = 0; nodes < nodes_per_fold; nodes++){
        outfile << "\nCONSTANT wr_tab_" << nodes+1 << " : wr_addr_list := ( ";
        outfile2 << "\nCONSTANT wr_tab_" << nodes+1 << " : wr_addr_list := ( ";
        for (int x = 0; x < (even_order*fold_factor)-1; x++){
            outfile << write_sch_mem_hyp[nodes][1][x] << ", ";
            outfile2 << write_sch_mem_poi[nodes][1][x] << ", ";
        }
        outfile << write_sch_mem_hyp[nodes][1][(even_order*fold_factor)-1] << " );";
        outfile2 << write_sch_mem_poi[nodes][1][(even_order*fold_factor)-1] << " );";
    }



    outfile << "\nBEGIN\n\n\ta1: addr1     <= \"" << default_addrZ << "\";\n\ta2: addr2     <= \"" << default_addrZ << "\";\n";
    outfile2 << "\nBEGIN\n\n\ta1: addr1     <= \"" << default_addrZ << "\";\n\ta2: addr2     <= \"" << default_addrZ << "\";\n";

    outfile.close();
    outfile2.close();

    outfile.open("Generated_Files/Portions/Wr_addrgen/wr_add3.vhd");
    outfile << "PROCESS( clk )\n";
    outfile << "    BEGIN\n";
    outfile << "        IF(clk'event and clk='1') THEN\n";
    outfile << "          IF en_addrgen /= '1' THEN -- Keep the counters in check.\n";
    outfile << "            addrcount <= 0;\n";
    outfile << "          ELSIF en_addrgen = '1' THEN\n            ";
    for (int nodes = 0; nodes < nodes_per_fold-1; nodes++){
        outfile << "IF( mu_id = \"" << binary(nodes, nodes_per_fold) << "\" ) THEN\n";
        outfile << "              addr1 <= CONV_STD_LOGIC_VECTOR( wr_tab_" << nodes+1 << "( addrcount ), " << mem_loc  << " );\n";
        outfile << "              addr2 <= CONV_STD_LOGIC_VECTOR( wr_tab_" << nodes+1 << "( addrcount + 1 ), " << mem_loc << " );\n";
        outfile << "            ELS";
    }
    outfile << "IF( mu_id = \"" << binary(nodes_per_fold - 1, nodes_per_fold) << "\" ) THEN\n";
    outfile << "              addr1 <= CONV_STD_LOGIC_VECTOR( wr_tab_" << nodes_per_fold << "( addrcount ), " << mem_loc << " );\n";
    outfile << "              addr2 <= CONV_STD_LOGIC_VECTOR( wr_tab_" << nodes_per_fold << "( addrcount + 1 ), " << mem_loc << " );\n";
    outfile << "            END IF;\n"    ;

    outfile << "            addrcount <= addrcount + 2;\n";
    outfile << "          END IF;\n";
    outfile << "        END IF;\n";
    outfile << "END PROCESS;\n\n";
    outfile << "END ARCHITECTURE behav;\n";
    outfile.close();

    outfile.open( outputfile1 );    // Creating the file outputfile1
    outfile.close();

    outfile.open( outputfile2 );    // Creating the file outputfile2
    outfile.close();

    append( "HDL_partitioned_files/Wr_addrgen/hyp_wr_add1.vhd", outputfile1 );
    append( "Generated_Files/Portions/Wr_addrgen/hyp_wr_add2.vhd", outputfile1 );
    append( "Generated_Files/Portions/Wr_addrgen/wr_add3.vhd", outputfile1 );

    append( "HDL_partitioned_files/Wr_addrgen/point_wr_add1.vhd", outputfile2 );
    append( "Generated_Files/Portions/Wr_addrgen/point_wr_add2.vhd", outputfile2 );
    append( "Generated_Files/Portions/Wr_addrgen/wr_add3.vhd", outputfile2 );
}

/* This function creates the Memory Module which contains the following
   entities:
        -- mem_unit
        -- tdp_mem
        -- rd_addrgen
        -- wr_addrgen
        -- add_mux                        */
void make_mem_module(const char * outputfile1, const char * outputfile2)
{

    ofstream outfile( outputfile1 );    // Creating the file outputfile1
    outfile.close();

    outfile.open( outputfile2 );    // Creating the file outputfile2
    outfile.close();

    // There are no changes to the top module of the memory unit.
    // So, the file is just picked up from HDL_Partitioned_files/Files_mem_module/mem_unit/
    // and appended to the right location in  .
    append( "HDL_partitioned_files/Mem_Unit/hyp_mem_unit.vhd", outputfile1 );
    append( "HDL_partitioned_files/Mem_Unit/point_mem_unit.vhd", outputfile2 );

    return;
}

