LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE WORK.fold_package.ALL;

ENTITY hyp_mem_unit IS
    PORT ( clk       : IN STD_LOGIC;
           rst       : IN STD_LOGIC;
           en_use    : IN STD_LOGIC;
           rd_w_bar1 : IN STD_LOGIC;
           rd_w_bar2 : IN STD_LOGIC;
           mu_id     : IN node_indx;
           d_in1     : IN STD_LOGIC;
           d_in2     : IN STD_LOGIC;
           d_out1    : OUT STD_LOGIC;
           d_out2    : OUT STD_LOGIC
         );
END ENTITY hyp_mem_unit;

ARCHITECTURE behav OF hyp_mem_unit IS
------------------address generator components----------------

COMPONENT rd_addrgen
        PORT ( clk          : IN STD_LOGIC;
               en_addrgen   : IN STD_LOGIC;
               addr1        : OUT mem_addr_bus;
               addr2        : OUT mem_addr_bus
             );
END COMPONENT;

COMPONENT hyp_wr_addrgen
        PORT ( clk          : IN STD_LOGIC;
               en_addrgen   : IN STD_LOGIC;
               mu_id        : IN node_indx;
               addr1        : OUT mem_addr_bus;
               addr2        : OUT mem_addr_bus
             );
END COMPONENT;

-------------------memory block component---------------------

COMPONENT tdp_mem
PORT (  clk     : IN STD_LOGIC;
        rst     : IN STD_LOGIC;
        en_mem  : IN STD_LOGIC;
        rwbar_1 : IN STD_LOGIC;
        rwbar_2 : IN STD_LOGIC;
        addr1   : IN mem_addr_bus;
        addr2   : IN mem_addr_bus;
        din_1   : IN STD_LOGIC;
        din_2   : IN STD_LOGIC;
        dout_1  : OUT STD_LOGIC;
        dout_2  : OUT STD_LOGIC
     );
END COMPONENT;

COMPONENT add_mux
    PORT ( clk          : IN STD_LOGIC;
           in_r_addr    : IN mem_addr_bus;
           in_w_addr    : IN mem_addr_bus;
           rd_w_bar     : IN STD_LOGIC;
           en           : IN STD_LOGIC;
           out_addr     : OUT mem_addr_bus
         );
END COMPONENT;

FOR ALL:rd_addrgen      USE ENTITY WORK.rd_addrgen( behav );
FOR ALL:hyp_wr_addrgen  USE ENTITY WORK.hyp_wr_addrgen( behav );
FOR ALL:add_mux         USE ENTITY WORK.add_mux( behav );
FOR ALL:tdp_mem         USE ENTITY WORK.tdp_mem( behav );

------------------------ internal signals --------------------------

SIGNAL  mem_rdwbar_1, mem_rdwbar_2          : STD_LOGIC;
SIGNAL  mux_rdwbar_1, mux_rdwbar_2          : STD_LOGIC;
SIGNAL  en_mem, en_mux, en_ruse, en_wuse    : STD_LOGIC;

SIGNAL  r_addr1, w_addr1, muxed_addr1  : mem_addr_bus;
SIGNAL  r_addr2, w_addr2, muxed_addr2  : mem_addr_bus;

BEGIN

a1: rd_addrgen  PORT MAP ( clk, en_ruse, r_addr1, r_addr2 );
a2: hyp_wr_addrgen  PORT MAP ( clk, en_wuse, mu_id, w_addr1, w_addr2 );

a3: add_mux     PORT MAP ( clk, r_addr1, w_addr1,
                            mux_rdwbar_1, en_mux, muxed_addr1  );

a4: add_mux     PORT MAP ( clk, r_addr2, w_addr2,
                            mux_rdwbar_2, en_mux, muxed_addr2 );

a5: tdp_mem     PORT MAP ( clk, rst, en_mem, mem_rdwbar_1, mem_rdwbar_2,
                                muxed_addr1, muxed_addr2,
                                        d_in1, d_in2, d_out1, d_out2 );

-- Another clause in the condition, AND rd_w_bar2 = '1' is not added
-- because some (last) perfect patterns may not use second port at times.
-- The following asynchronous assignment saves one cycle, especially in
-- microcode-sequenced hardware.
a6: en_ruse <= '1' WHEN ( en_use = '1' AND rd_w_bar1 = '1' ) ELSE '0';
a7: en_wuse <= '1' WHEN ( en_use = '1' AND rd_w_bar1 = '0' ) ELSE '0';

-- The mux is enabled one cycle later than address generation enable, so
-- that muxing starts as soon as address become available.
PROCESS( clk )
BEGIN
    IF (clk'event AND clk='1') THEN
      IF( en_use = '1' ) THEN
        en_mux <= '1';
      ELSE
        en_mux <= '0';
      END IF;
   END IF;
END PROCESS;

-- One cycle delay of mux enable. Since mux is negative edge-triggered, in
-- the next/second cycle, read/write can ensue from the actual memory.
PROCESS( clk )
BEGIN
    IF (clk'event AND clk='1') THEN
      mux_rdwbar_1 <= rd_w_bar1;
      mux_rdwbar_2 <= rd_w_bar2;
   END IF;
END PROCESS;

-- One cycle delayed reading from actual memory. The reading happens one
-- cycle later than preparation of the addresses.
PROCESS( clk )
BEGIN
    IF (clk'event AND clk='1') THEN
      en_mem <= en_use;
      mem_rdwbar_1 <= rd_w_bar1;
      mem_rdwbar_2 <= rd_w_bar2;
   END IF;
END PROCESS;

END ARCHITECTURE behav;
