LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.STD_LOGIC_UNSIGNED.ALL;
USE WORK.fold_package.ALL;

ENTITY bit_f_unit IS
    PORT( clk            : IN STD_LOGIC;
          in_bit1        : IN STD_LOGIC;
          in_bit2        : IN STD_LOGIC;
          intrinsic_vec  : IN bus_instr_vec;
          fold_indx      : IN fold_indx;
          en_acc         : IN STD_LOGIC;
          en_flip        : IN STD_LOGIC;
          en_first_time  : IN STD_LOGIC;
          en_part_store  : IN STD_LOGIC;
          en_count_reset : IN STD_LOGIC;
          en_out         : IN STD_LOGIC;
          bit_estimate   : OUT STD_LOGIC;
          out_bit1       : OUT STD_LOGIC;
          out_bit2       : OUT STD_LOGIC
        );
END ENTITY bit_f_unit;

ARCHITECTURE behav OF bit_f_unit IS
    SIGNAL no_zeroes_0, no_zeroes_1, no_zeroes_2     : INTEGER;
    SIGNAL no_ones_0, no_ones_1, no_ones_2     : INTEGER;
    SIGNAL bit_estimate_0, bit_estimate_1, bit_estimate_2     : STD_LOGIC;
BEGIN

-- Counts the success and failure of parity checks, 2 check results at a
-- time.
PROCESS( clk )
BEGIN
    IF(clk'event and clk='1') THEN
      IF( en_acc = '1' ) THEN
          -- Aggregation is used since two concurrent increments of a
          -- counter value does NOT lead to cumulative increase in value of
          -- the counter by 2, which is desired for certain input patterns.
        IF( fold_indx = "00" ) THEN
          IF( ( in_bit1, in_bit2 ) = "11" ) THEN
            no_ones_0     <= no_ones_0 + 2;
          ELSIF( ( in_bit1, in_bit2 ) = "10" ) THEN
            no_zeroes_0   <= no_zeroes_0 + 1;
            no_ones_0     <= no_ones_0 + 1;
          ELSIF( ( in_bit1, in_bit2 ) = "01" ) THEN
            no_zeroes_0   <= no_zeroes_0 + 1;
            no_ones_0     <= no_ones_0 + 1;
          ELSIF( ( in_bit1, in_bit2 ) = "00" ) THEN
            no_zeroes_0   <= no_zeroes_0 + 2;
          ELSE
            IF( in_bit1 = '1' ) THEN -- in_bit2 may be Z, X etc.
              no_ones_0   <= no_ones_0 + 1;
            ELSIF( in_bit1 = '0' ) THEN -- in_bit2 may be Z, X etc.
              no_zeroes_0 <= no_zeroes_0 + 1;
            END IF;

            IF( in_bit2 = '1' ) THEN -- in_bit1 may be Z, X etc.
              no_ones_0 <= no_ones_0 + 1;
            ELSIF( in_bit2 = '0' ) THEN -- in_bit1 may be Z, X etc.
              no_zeroes_0 <= no_zeroes_0 + 1;
            END IF;
          END IF;
        ELSIF( fold_indx = "01" ) THEN
          IF( ( in_bit1, in_bit2 ) = "11" ) THEN
            no_ones_1     <= no_ones_1 + 2;
          ELSIF( ( in_bit1, in_bit2 ) = "10" ) THEN
            no_zeroes_1   <= no_zeroes_1 + 1;
            no_ones_1     <= no_ones_1 + 1;
          ELSIF( ( in_bit1, in_bit2 ) = "01" ) THEN
            no_zeroes_1   <= no_zeroes_1 + 1;
            no_ones_1     <= no_ones_1 + 1;
          ELSIF( ( in_bit1, in_bit2 ) = "00" ) THEN
            no_zeroes_1   <= no_zeroes_1 + 2;
          ELSE
            IF( in_bit1 = '1' ) THEN -- in_bit2 may be Z, X etc.
              no_ones_1   <= no_ones_1 + 1;
            ELSIF( in_bit1 = '0' ) THEN -- in_bit2 may be Z, X etc.
              no_zeroes_1 <= no_zeroes_1 + 1;
            END IF;

            IF( in_bit2 = '1' ) THEN -- in_bit1 may be Z, X etc.
              no_ones_1 <= no_ones_1 + 1;
            ELSIF( in_bit2 = '0' ) THEN -- in_bit1 may be Z, X etc.
              no_zeroes_1 <= no_zeroes_1 + 1;
            END IF;
          END IF;
        ELSIF( fold_indx = "10" ) THEN
          IF( ( in_bit1, in_bit2 ) = "11" ) THEN
            no_ones_2     <= no_ones_2 + 2;
          ELSIF( ( in_bit1, in_bit2 ) = "10" ) THEN
            no_zeroes_2   <= no_zeroes_2 + 1;
            no_ones_2     <= no_ones_2 + 1;
          ELSIF( ( in_bit1, in_bit2 ) = "01" ) THEN
            no_zeroes_2   <= no_zeroes_2 + 1;
            no_ones_2     <= no_ones_2 + 1;
          ELSIF( ( in_bit1, in_bit2 ) = "00" ) THEN
            no_zeroes_2   <= no_zeroes_2 + 2;
          ELSE
            IF( in_bit1 = '1' ) THEN -- in_bit2 may be Z, X etc.
              no_ones_2   <= no_ones_2 + 1;
            ELSIF( in_bit1 = '0' ) THEN -- in_bit2 may be Z, X etc.
              no_zeroes_2 <= no_zeroes_2 + 1;
            END IF;

            IF( in_bit2 = '1' ) THEN -- in_bit1 may be Z, X etc.
              no_ones_2 <= no_ones_2 + 1;
            ELSIF( in_bit2 = '0' ) THEN -- in_bit1 may be Z, X etc.
              no_zeroes_2 <= no_zeroes_2 + 1;
            END IF;
          END IF;
        END IF;
      ELSE
        -- Counters need to be reset, every iteration. The bit
        -- estimate needs to be preserved, though. Hence a
        -- component-wide reset cannot be used.
        IF( en_count_reset = '1' ) THEN
            no_ones_0    <= 0;
            no_zeroes_0    <= 0;
            no_ones_1    <= 0;
            no_zeroes_1    <= 0;
            no_ones_2    <= 0;
            no_zeroes_2    <= 0;
        END IF;
      END IF;
    END IF;
END PROCESS;

-- The 'bit-flipping' process, based on what was in majority: parity check
-- success, or parity check failure. It calculates the estimate of a bit,
-- every iteration, as a byproduct.
PROCESS( clk )
BEGIN
    IF(clk'event AND clk='1') THEN
      -- Convert the intrinsic information into the first-ever bit estimate.
      IF( en_first_time = '1' ) THEN
        IF( fold_indx = "00" ) THEN
          IF( intrinsic_vec(8) = '1' ) THEN -- negative value
            bit_estimate_0 <= '1';
          ELSIF( intrinsic_vec(8) = '0' ) THEN -- positive value
            bit_estimate_0 <= '0';
          END IF;
        ELSIF( fold_indx = "01" ) THEN
          IF( intrinsic_vec(8) = '1' ) THEN -- negative value
            bit_estimate_1 <= '1';
          ELSIF( intrinsic_vec(8) = '0' ) THEN -- positive value
            bit_estimate_1 <= '0';
          END IF;
        ELSIF( fold_indx = "10" ) THEN
          IF( intrinsic_vec(8) = '1' ) THEN -- negative value
            bit_estimate_2 <= '1';
          ELSIF( intrinsic_vec(8) = '0' ) THEN -- positive value
            bit_estimate_2 <= '0';
          END IF;
        END IF;

      -- Keep on updating the bit estimate.
      -- IF there are more failed parity checks than succeeded parity
      -- checks, toggle the bit estimate. Otherwise, even if they are
      -- equal, toggle them. This way, in case there is tie at every bit,
      -- the opposite bit pattern can be tried out for parity check
      -- success. Else, if all bits remain what they were, the failure will
      -- continue ad infinitum.
      ELSIF( en_flip = '1' ) THEN
        IF( fold_indx = "00" ) THEN
          IF( no_ones_0 >= no_zeroes_0 ) THEN -- more parity check failures
            bit_estimate_0 <= NOT( bit_estimate_0 );
          ELSE
            bit_estimate_0 <= bit_estimate_0;
          END IF;
        ELSIF( fold_indx = "01" ) THEN
          IF( no_ones_1 >= no_zeroes_1 ) THEN -- more parity check failures
            bit_estimate_1 <= NOT( bit_estimate_1 );
          ELSE
            bit_estimate_1 <= bit_estimate_1;
          END IF;
        ELSIF( fold_indx = "10" ) THEN
          IF( no_ones_2 >= no_zeroes_2 ) THEN -- more parity check failures
            bit_estimate_2 <= NOT( bit_estimate_2 );
          ELSE
            bit_estimate_2 <= bit_estimate_2;
          END IF;
        END IF;
      END IF;
    END IF;

END PROCESS;

-- Switch/output the right bit estimate, in right clock cycle, for
-- consumption of check nodes(2 at a time).
PROCESS( clk )
BEGIN
    IF(clk'event and clk='1') THEN
      IF( en_out = '1' ) THEN
        IF( fold_indx = "00" ) THEN
          bit_estimate  <= bit_estimate_0;
          out_bit1      <= bit_estimate_0;
          out_bit2      <= bit_estimate_0;
        ELSIF( fold_indx = "01" ) THEN
          bit_estimate  <= bit_estimate_1;
          out_bit1      <= bit_estimate_1;
          out_bit2      <= bit_estimate_1;
        ELSIF( fold_indx = "10" ) THEN
          bit_estimate  <= bit_estimate_2;
          out_bit1      <= bit_estimate_2;
          out_bit2      <= bit_estimate_2;
        END IF;
      END IF;
    END IF;

END PROCESS;

END ARCHITECTURE behav;
